/**
 * Hlavní semestrální práce z KIV/VSP
 * (c) 2006 - 2007 Ivan Habernal
 * e-mail: habernal@students.zcu.cz
 */
package cz.habi.vsp.sem.network;

import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.commons.math.stat.descriptive.SummaryStatistics;

import cz.habi.vsp.sem.distribution.Distribution;
import cz.zcu.fav.kiv.jsim.JSimException;

/**
 * This class represents a server. It accepts the requests, holds them in the
 * waiting queue (if needed) and process them, which costs some time.
 * 
 * @author habi
 * 
 */
public class Server extends NetworkProcess implements RequestAcceptor, StatisticsProvider {

	private Queue queue;

	private boolean idle = true;

	// inner server statistic - how many requests have been served
	private int servedRequests = 0;

	// Tq statistics
	private SummaryStatistics tqStatistics = SummaryStatistics.newInstance();

	// time, when the server was busy (= working)
	private double totalBusyTime = 0.0;

	private double startingTime;

	/**
	 * Creates a new instance of Server
	 * 
	 * @param name
	 *            name
	 * @param simulation
	 *            parent simulation
	 * @param distribution
	 *            probability distribution
	 * @throws JSimException
	 *             exception
	 * @throws IllegalArgumentException
	 *             if the distribution is null
	 */
	public Server(String name, SystemSimulation simulation, Distribution distribution) throws JSimException, IllegalArgumentException {
		super(name, simulation, distribution);
		queue = new Queue(simulation, this);
		this.startingTime = simulation.getCurrentTime();
	}

	public void acceptRequest(Request request) throws JSimException {
		// every request must wait in the queue
		queue.addRequestToQueue(request);

		// if the server is idle, we wake him up
		if (idle) {
			this.activateNow();
		}
	}

	@Override
	protected void doWork() throws JSimException {
		Request request = queue.getNextRequest();

		// we have nothing to do -> sleep
		if (request == null) {
			idle = true;
			this.passivate();
		} else {
			idle = false;

			// "working" for some time
			double busyTime = holdForRandomTime();
			this.totalBusyTime += busyTime;

			// we served one request
			servedRequests++;

			// updating Tq statistics (time spent in queue and server)
			tqStatistics.addValue(request.getWaitingTimeInLastQueue() + busyTime);

			// and pass the request to the next element
			getNextRequestAcceptor().acceptRequest(request);
		}
	}

	@Override
	protected boolean workingCondition() {
		// Servers just work forever...
		return true;
	}

	/**
	 * Returns the average load of the server.
	 * 
	 * @return load
	 */
	public double getLoad() {
		double result = 0.0;

		double currentTime = getParent().getCurrentTime();
		if (currentTime > 0) {
			result = this.totalBusyTime / currentTime;
		}

		return result;
	}

	/**
	 * Returns the T<sub>Q</sub> - average time of passing through the queue
	 * and the server.
	 * 
	 * @return the average time
	 */
	public double getAverageResponseTime() {
		return this.tqStatistics.getMean();
	}

	/**
	 * Returns the standard deviation of T<sub>Q</sub>.
	 * 
	 * @return standard deviation
	 */
	public double getDeviationResponseTime() {
		return this.tqStatistics.getStandardDeviation();
	}

	public Map<String, Object> getStatistics() {
		Map<String, Object> map = new LinkedHashMap<String, Object>();

		map.put("Load", getLoad());
		map.put("Tq", getAverageResponseTime());
		map.put("Tq (DEV)", getDeviationResponseTime());
		map.put("Requests", this.servedRequests);
		map.put("Ts", totalBusyTime / (double) servedRequests);
		map.put("Flow", servedRequests / (getParent().getCurrentTime() - startingTime));

		return map;
	}

	public String getProviderName() {
		return "(Server): " + getProcessName();
	};

	/**
	 * Returns queue statistics.
	 * 
	 * @return queue stats.
	 */
	public StatisticsProvider getQueueStatistics() {
		return queue;
	}
}
