/**
 * Hlavní semestrální práce z KIV/VSP
 * (c) 2006 - 2007 Ivan Habernal
 * e-mail: habernal@students.zcu.cz
 */
package cz.habi.vsp.sem.network;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.apache.commons.math.stat.descriptive.SummaryStatistics;

import cz.zcu.fav.kiv.jsim.JSimException;
import cz.zcu.fav.kiv.jsim.JSimInvalidParametersException;
import cz.zcu.fav.kiv.jsim.JSimSimulation;

/**
 * This is the main simulation for queuing networks. It provides also some
 * staticstics.
 * 
 * @author habi
 * 
 */
public class SystemSimulation extends JSimSimulation implements StatisticsProvider {
	private SummaryStatistics tqStats = SummaryStatistics.newInstance();

	// count of all requests present in the system
	private int actualRequestsCount = 0;

	// how many requests must go through the system
	private int totalRequestsCountToPass;

	// how many request has already passed through
	private int totalPassedRequests = 0;

	// steps
	private int stepsCount = 0;

	// starting time
	private double startingTime;

	// shutdown time
	private double shutdownTime;
	
	private double lqTemp = 0;

	private double lastRequestsCountChangeTime;

	// set of statistics providers
	private Set<StatisticsProvider> statisticsProviders = new LinkedHashSet<StatisticsProvider>();

	/**
	 * Creates a new instance of SystemSimulation
	 * 
	 * @param name
	 *            name
	 * @param totalRequestsCount
	 *            total requests
	 * @throws JSimInvalidParametersException
	 *             if an exception occurs
	 * @throws IllegalArgumentException
	 *             if totalRequestsCount is zero or negative
	 */
	public SystemSimulation(String name, int totalRequestsCount) throws JSimInvalidParametersException, IllegalArgumentException {
		super(name);

		if (totalRequestsCount <= 0) {
			throw new IllegalArgumentException("RequestsCount must be positive");
		}
		this.totalRequestsCountToPass = totalRequestsCount;

		this.startingTime = this.getCurrentTime();
		this.lastRequestsCountChangeTime = startingTime;
		
		// adding itself to statistics providers
		registerStatisticsProvider(this);
	}

	/**
	 * Runs the system simulation.
	 * 
	 * @throws JSimException
	 *             if an error occurs
	 */
	public void runSimulation() throws JSimException {
		boolean working = true;

		while (working) {
			boolean step = this.step();

			if (step) {
				stepsCount++;
			}

			working &= step && (totalPassedRequests <= totalRequestsCountToPass);
		}
		
		shutdownTime = getCurrentTime();

		shutdown();
	}

	/**
	 * Return the steps count of simulation
	 * 
	 * @return steps count
	 */
	public int getStepsCount() {
		return stepsCount;

	}

	/**
	 * Updates Lq statistics
	 */
	private void updateLqStatistics() {
		double interval = this.getCurrentTime() - lastRequestsCountChangeTime;
		lqTemp += interval * (double) actualRequestsCount;

		lastRequestsCountChangeTime = this.getCurrentTime();
	}

	/**
	 * Must be called, when a request is created.
	 */
	public synchronized void requestEnteredToSystem() {
		updateLqStatistics();

		actualRequestsCount++;
	}

	/**
	 * Must be called, when a request leaves the system.
	 */
	public synchronized void requestLeftSystem() {
		updateLqStatistics();

		actualRequestsCount--;

		if (actualRequestsCount < 0) {
			throw new IllegalStateException("System contains negative count of requests!");
		}

		totalPassedRequests++;
	}

	/**
	 * Adds a value to the Tq statistics
	 * 
	 * @param value
	 */
	public void addTqStatsValue(double value) {
		tqStats.addValue(value);
	}

	/**
	 * Returns the T<sub>Q</sub> - average time of passing through the system.
	 * 
	 * @return the average time
	 */
	public double getTq() {
		return tqStats.getMean();
	}

	/**
	 * Returns the standard deviation of T<sub>Q</sub>.
	 * 
	 * @return standard devaition
	 */
	public double getDeviationTq() {
		return tqStats.getStandardDeviation();
	}

	/**
	 * Returns the L<sub>Q</sub> - average count of requests in the system.
	 * 
	 * @return average count of requests
	 */
	public double getLq() {
		return lqTemp / (getCurrentTime() - startingTime);
	}

	public String getProviderName() {
		return "(System): " + getSimulationName();
	}

	public Map<String, Object> getStatistics() {
		Map<String, Object> result = new LinkedHashMap<String, Object>();

		result.put("Stopped at", shutdownTime);
		result.put("Requests", totalPassedRequests);
		result.put("Steps", getStepsCount());
		result.put("Tq", getTq());
		result.put("Tq (DEV)", getDeviationTq());
		result.put("Lq", getLq());

		return result;
	}

	/**
	 * Registers an instance of {@link StatisticsProvider} to the system. The
	 * provider is class, which provides some statistics information.
	 * 
	 * @param provider
	 *            provider
	 * @throws IllegalArgumentException
	 *             if the provider is null
	 */
	public void registerStatisticsProvider(StatisticsProvider provider) throws IllegalArgumentException {
		statisticsProviders.add(provider);
	}

	/**
	 * Returns a set of all registered statistics providers.
	 * 
	 * @return set of statistics providers
	 */
	public Set<StatisticsProvider> getStatisticProviders() {
		return statisticsProviders;
	}

}
