/**
 * Hlavní semestrální práce z KIV/VSP
 * (c) 2006 - 2007 Ivan Habernal
 * e-mail: habernal@students.zcu.cz
 */
package cz.habi.vsp.sem.tests;

import static org.junit.Assert.assertTrue;
import junit.framework.JUnit4TestAdapter;

import org.apache.commons.math.stat.descriptive.SummaryStatistics;
import org.junit.Test;

import cz.habi.vsp.sem.distribution.Distribution;
import cz.habi.vsp.sem.distribution.ErlangDistribution;
import cz.habi.vsp.sem.distribution.ExponentialDistribution;
import cz.habi.vsp.sem.distribution.NormalDistribution;
import cz.habi.vsp.sem.distribution.UniformDistribution;

/**
 * This test case covers some distributution classes and its random number
 * generators.
 * 
 * @author habi
 * 
 */
public class DistributionsTest {
	
	private static final boolean ALWAYS_SHOW_RESULTS = false;

	private static final int NUMBER_OF_REPEATS = 1000000;

	private static final double EPSILON = 0.05;

	/**
	 * Runs the generator for NUMBER_OF_REPEATS times and returns the
	 * statistics.
	 * 
	 * @param distribution
	 *            distribution
	 * @return array of double, where result[0] = mean and result[1] = variance
	 */
	private final double[] runGenerator(Distribution distribution) {
		double[] result = new double[2];
		SummaryStatistics stats = SummaryStatistics.newInstance();

		for (int i = 0; i < NUMBER_OF_REPEATS; i++) {
			stats.addValue(distribution.getRandomValue());
		}

		result[0] = stats.getMean();
		result[1] = stats.getVariance();

		return result;
	}

	/**
	 * Returns true, if the given distribution's simulated mean and variance are
	 * equal to theoretic mean and variance. The equation is like following:
	 * 
	 * <pre>
	 * abs(sim.value - theor.value) &lt; EPSILON
	 * </pre>
	 * 
	 * @param distribution
	 *            distribution
	 * @return boolean value
	 */
	private final boolean internalTestDistribution(Distribution distribution) {
		double[] results = runGenerator(distribution);

		double meanDiff = Math.abs(distribution.getMean() - results[0]);
		double varianceDiff = Math.abs(distribution.getVariance() - results[1]);

		boolean result = (meanDiff < EPSILON) && (varianceDiff < EPSILON);

		if (!result || ALWAYS_SHOW_RESULTS) {
			System.out.println("+===========================================");
			System.out.println("| Testing " + distribution.getClass().getName());
			System.out.println("+===========================================");
			System.out.println("| Mean:");
			System.out.println("|   theory: " + distribution.getMean() + ", simulated: " + results[0]);
			System.out.println("| Variance:");
			System.out.println("|   theory: " + distribution.getVariance() + ", simulated: " + results[1]);
			System.out.println("| Difference:");
			System.out.println("|   meanDiff     = " + meanDiff);
			System.out.println("|   varianceDiff = " + varianceDiff);
			System.out.println("+-------------------------------------------");
		}

		return result;
	}

	/**
	 * Tests the function of exponential distribution class.
	 */
	@Test
	public final void testExponentialDistribution() {
		Distribution distribution = new ExponentialDistribution(5.0);
		assertTrue(internalTestDistribution(distribution));

		distribution = new ExponentialDistribution(1.0);
		assertTrue(internalTestDistribution(distribution));
		
		distribution = new ExponentialDistribution(1000000.0);
		assertTrue(internalTestDistribution(distribution));
	}

	/**
	 * Tests the function of normal distribution class.
	 */
	@Test
	public final void testNormalDistribution() {
		Distribution d = new NormalDistribution(5.0, 2.0);
		assertTrue(internalTestDistribution(d));
		
		d = new NormalDistribution(0.0, 0.0);
		assertTrue(internalTestDistribution(d));
		
		d = new NormalDistribution(100000.0, 0.5);
		assertTrue(internalTestDistribution(d));
	}

	/**
	 * Tests the function of erlang distribution class.
	 */
	@Test
	public final void testErlangDistribution() {
		Distribution d = new ErlangDistribution(1.0);
		assertTrue(internalTestDistribution(d));
		
		d = new ErlangDistribution(100000.0);
		assertTrue(internalTestDistribution(d));
	}

	/**
	 * Tests the function of uniform distribution
	 */
	@Test
	public final void testUniformDistribution() {
		Distribution d = new UniformDistribution(-1, 1);
		assertTrue(internalTestDistribution(d));
		
		d = new UniformDistribution(0, 10);
		assertTrue(internalTestDistribution(d));
	}

	public static junit.framework.Test suite() {
		return new JUnit4TestAdapter(DistributionsTest.class);
	}
}
